/*
	Ac3dec plugin for SoundPlay by Jerome Duval 09/2000
	Original free sources 
*/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <AppKit.h>
#include <InterfaceKit.h>
#include <MediaDefs.h>
#include <Path.h>
#include "Ac3decPlugin.h"
#include "ac3decode.h"
#include "Preferences.h"
#include "PrefsPanel/AC3EdMainWindow.h"

// raw configuration
#define NCH 2
#define SAMPLERATE samplerate
#define BPS 16
#define NSAMPS	(256*6)

extern long SampleRate;
extern long BitRate;

char lastfn[256]; // currently playing file (used for getting info on the current file)
int file_length; // file length, in bytes
int decode_pos_ms; // current decoding position, in milliseconds
int play_pos_ms; // current playing position, in milliseconds
int paused; // are we paused?
int seek_needed; // if != -1, it is the point that the decode thread should seek to, in ms.
static char sample_buffer[NSAMPS*NCH*(BPS/8)*2]; // sample buffer
const int buffer_size = NSAMPS*NCH*(BPS/8);
const int frame_size = NCH * BPS / 8;
int buffer_time = 0;
long samplerate = 48000;
long TotalTime = 0;
sem_id write_sem, read_sem;
int killDecodeThread=0;
bool end_of_file = false;

extern FILE *g_pInfile;

int32 DecodeThread(void *b);

long GetFileLength( char *fileName )
{
	FILE 	*fp;
	long	len=0;
	
	if ( (fp = fopen( fileName, "rb" )) ) {
		fseek( fp, 0, SEEK_END );
		len = ftell( fp );
		fseek( fp, 0, SEEK_SET );
		fclose( fp );
	}
	return len;
}


/* ===========================================================
	Plugin-descriptor structure.
	This structure is used to gather information about a plugin
	and	instantiate and destroy specific instances of a plugin.
*/
static plugin_descriptor plugin={
	PLUGIN_DESCRIPTOR_MAGIC,PLUGIN_DESCRIPTOR_VERSION,
	"ac3dec",1, PLUGIN_IS_DECODER,

	"Ac3dec player",
	"By Jerome Duval.\n\n"
	"Based on Ac3dec free sources.\n",
	&aboutplugin,
	&configure,
	&getplugin,
	&destroyplugin
};

/* ===========================================================
	Plugin-operations structure.
	This is used to actually perform operations on a plugin.
	It is returned by the getplugin() function from the 
	plugin_descriptor structure above.
*/
static decoder_plugin_ops ops={
	PLUGIN_DECODER_MAGIC,PLUGIN_DECODER_VERSION,

	open,
	close,
	info,
	readdata,
	NULL,	// play
	NULL,	// stop
	NULL,	// setvolume
	NULL,	// setpitchandspeed
	seek,
	position,
	NULL  // bufferamount
};


/* ===========================================================
	Return the list of plugins in this file.
*/
static plugin_descriptor *plugs[]={
	&plugin,
	0
};

plugin_descriptor **get_plugin_list(void)
{
	return plugs;
}

/* ===========================================================
	Return the list of filetypes supported by all the plugins
	in this file
*/
static supported_type supportedtypes[]=
{
	{"audio/x-ac3",		"Dolby Ac3 Audio File",							"Ac3 File",	"ac3", NULL },
	{NULL,				NULL,															NULL,						NULL,NULL}
};

supported_type *get_supported_types(void)
{
	return supportedtypes;
}


/* ===========================================================
	Implementation of the plugin_descriptor functions
*/
static bool mmRegistered=false;
static void *getplugin(void **data, const char *name, const char *header, uint32 size, plugin_info *info)
{
	if(!mmRegistered)
	{
		mmRegistered=true;
	
	}

	if(info)
	{
		// Loading a new instance of this plugin is an easy way
		// around rewriting everything to use local variables.
		// Using the "main" image as well instead of always
		// loading a new image is left as an excercise for the reader.
		BEntry entry(info->ref);
		if(!entry.Exists())
			return NULL;
		BPath path(&entry);
		image_id img=load_add_on(path.Path());
		if(img<0)
			return NULL;
		plugin_descriptor** (*declistfunc)(void);
		if(B_OK!=get_image_symbol(img, "get_plugin_list", B_SYMBOL_TYPE_TEXT, (void**)&declistfunc))
			return NULL;
		plugin_descriptor *plugdesc=*(declistfunc());
		void *imgops=plugdesc->Instantiate_Plugin(data,name,header,size,NULL);
		if(imgops)
		{
			ac3_handle *ah=(ac3_handle*)*data;
			ah->img=img;
			ah->plugdesc=plugdesc;
			ah->prefs = ThePrefs;
			return imgops;
		}
		unload_add_on(img);
	}
	else
	{
		ac3_handle *ah=new ac3_handle;
		char *ext=strrchr(name,'.');
		// Check if possible Url
		bool http= (0==strncasecmp(name,"http://",7));
		// In this example, we don't want Url's
		if (!http && ext) {
			if((0==strcasecmp(ext,".ac3"))||(0==strcasecmp(ext,".vob"))) {
				ah->af = OpenAc3((char*)name);
				*data=(void*)ah;
				return &ops;
				}
		}
		delete ah;		
	}
	return NULL;
}

static void destroyplugin(void *ops,void *_ah)
{
	ac3_handle *ah=(ac3_handle*)_ah;
	if(ops)
	{
		ah->plugdesc->Destroy_Plugin(NULL,_ah);
		unload_add_on(ah->img);
		
		delete ah;
		return;
	}
}

// Normally you wouldn't implement the about function for such a simple
// message. It is done here for completeness.
static void aboutplugin()
{
	BAlert *alert=new BAlert("About Ac3dec player",
		"Ac3dec player.\n"
		"By Jerome Duval.\n\n"
		"Based on Ac3dec free sources.\n",
		"OK");
	alert->SetFeel(B_NORMAL_WINDOW_FEEL);
	alert->Go(NULL);
}

BView* configure(BMessage *config)
{
	BRect rect(10, 20, 250, 340);
	// creating main view 
	AC3EdView_L0S0 *view_L0S0 = new AC3EdView_L0S0(rect, "view_L0S0", B_FOLLOW_LEFT, B_WILL_DRAW);

    // specifying background color for view_L0S0
	rgb_color color = {216,216,216,0};
	view_L0S0->SetViewColor(color);
	
	//return ;
	return view_L0S0;
}

status_t open(void* _ah,const char *name, const char *header, uint32 size, BMessage *config)
{
	
	ac3_handle *ah=(ac3_handle*)_ah;

	end_of_file = false;
	seek_needed = -1;
	
	file_length = GetFileLength((char*)name);

	if ((write_sem = create_sem(1, "write")) < B_NO_ERROR) 
		return B_OK;
	if ((read_sem = create_sem(0, "read")) < B_NO_ERROR) {
		delete_sem(write_sem);
		return B_OK;
	}

	if (ah->prefs)
 			ThePrefs = (Preferences*)ah->prefs;

	killDecodeThread=0;
	ah->dec_id = spawn_thread(DecodeThread,"Ac3decodeThread", 80,(void *) &killDecodeThread);

	if ((resume_thread(ah->dec_id))== B_OK) {
		buffer_time = (NSAMPS*1000)/SampleRate;
		return B_OK;
	}
	else
	{	
		return B_ERROR;	
	}
}

void close(void* _ah)
{
		
	killDecodeThread=1;
	
	if (g_pInfile != 0)
	{
		CloseAc3();
		g_pInfile = 0;
	}
	
	//kill_thread(ah->dec_id);
	
	delete_sem(write_sem);
	delete_sem(read_sem);
	
}

/* ===========================================================
	Implementation of the decoder_plugin_ops functions
*/
int32 readdata(void *_ah, char *buf, ulong count)
{
	ac3_handle *ah=(ac3_handle*)_ah;
	
	
	int Cnt= 0; // Returning 0 means End Of File
	
	if (acquire_sem(read_sem) != B_NO_ERROR)
		return B_ERROR;
	
	play_pos_ms = decode_pos_ms;
			
	memcpy(buf, sample_buffer, buffer_size);
	
	if (!end_of_file)
		Cnt = buffer_size/frame_size; //nb of frames
	else
		Cnt = 0;
			
	release_sem(write_sem);
	return (Cnt);
}

status_t info(void *_ah, file_info *info)
{
	ac3_handle *ah=(ac3_handle*)_ah;

// Fill in the file_info structure
	strcpy(info->name,"");
	strcpy(info->typedesc,"Dolby Ac3 File");
	strcpy(info->mimetype,"audio/x-ac3"); 
	info->samplerate=SampleRate;
	info->bitrate=BitRate*1000;
	
	TotalTime = file_length*8/BitRate;
	info->numchannels=2;
	info->granularity=1536;
	info->framecount=TotalTime*(SampleRate/1000);
	info->samplesize=2;
	if(B_HOST_IS_LENDIAN)
		info->byteorder=B_LITTLE_ENDIAN;
	else
		info->byteorder=B_BIG_ENDIAN;
	info->sampleformat=B_LINEAR_SAMPLES;
	info->flags=PLUGIN_NO_ELAPSEDTIME;
	
	return B_OK;
}

static status_t	seek(void *_ah, uint32 pos)
{
	
	ac3_handle *ah=(ac3_handle*)_ah;
	
	if (acquire_sem(read_sem) != B_NO_ERROR)
		return B_ERROR;
		
	seek_needed = pos/SampleRate*1000;
	
	release_sem(write_sem); 
	
	return B_OK;
}

static uint32 position(void *_ah)
{
	ac3_handle *ah=(ac3_handle*)_ah;
	uint32 pos;
	
	pos = play_pos_ms*(SampleRate/1000);
	
	return pos;
}

int main()
{
	BApplication app("application/x-jduval-ac3plugin");
	BAlert *alert=new BAlert("About Ac3dec player",
		"Ac3dec plugin for SoundPlay.\n"
		"By Jerome Duval.\n\n"
		"Based on Ac3dec free sources.\n",
		"OK");
	alert->SetFeel(B_NORMAL_WINDOW_FEEL);
	alert->Go();
}

int32 DecodeThread(void *b)
{
	
	while (! *((int *)b) && !end_of_file) 
	{
			
		if (acquire_sem(write_sem) != B_NO_ERROR)
			return B_ERROR;
			
		if (seek_needed != -1) {
		    if (seek_needed <= TotalTime){
				if ((seek_needed -=  3*buffer_time) <0)
					seek_needed = 0;
				int32 offs = (int64)seek_needed*((int64)file_length)/(int64)TotalTime;
			
				fseek( g_pInfile, offs, SEEK_SET);
				decode_pos_ms = seek_needed;
				seek_needed=-1;
			
				if (DecodeAc3(sample_buffer))
					end_of_file = true;
				decode_pos_ms += buffer_time;
				if (DecodeAc3(sample_buffer))
					end_of_file = true;
				decode_pos_ms += buffer_time;
				if (DecodeAc3(sample_buffer))
					end_of_file = true;
				decode_pos_ms += buffer_time;
				if (DecodeAc3(sample_buffer))
					end_of_file = true;
			}
			else {
				decode_pos_ms = seek_needed;
				seek_needed = -1;
			}
		}
		else
		if (DecodeAc3(sample_buffer))
			end_of_file = true;
		
		decode_pos_ms += buffer_time;
		
		release_sem(read_sem);
		
	}
	return 0;
}


